
/*- 
 * 
 * New BSD License 2006
 *
 * Copyright (c) 2006, Jorgen Lundman
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1 Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 2 Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.  
 * 3 Neither the name of the stuff nor the names of its contributors 
 *   may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

// $Id: pipe_win32.c,v 1.17 2006/06/30 01:22:58 lundman Exp $
// Win32 specific solutions to help with pipes.
// Jorgen Lundman January 29th, 2003.

#ifdef WIN32

// Universal system headers
#define _BSD_SIGNALS
#include <stdio.h>
#define WINDOWS_LEAN_AND_MEAN
#include <winsock2.h>
#include <io.h>
#include <errno.h>

#include "connections.h"
#include "lion.h"
#include "io.h"
#include "pipe.h"
#include "sockets.h"
#include "misc.h"


// Right, here's the deal. We want to handle sockets, and files with one
// place of sleep/cpu-release. to be maximum efficient.
//
// But Windows don't allow file handles to be added to select().
// So we try WaitForMultipleObjects, and Overlapped IO
// but Win 95, 98 and ME don't handle Overlapped IO on disk files.
// There is no fcntl for nonblocking under windows.
// Right, seems the answer in Windows is to thread! Rather inefficient
// but if a user is running Windows they can't be all that concerned about
// efficiency.
// So we thread and create a pipe. But anonymous pipes are uni-directional.
// So we have to create a named pipe. 
// Named pipes are not supported under Win95, 98 and ME.
// So, lets call socketpair instead.
// There is no socketpair under winsock.
// So lets manually create a socketpair, thread off to deal with 
// file IO to is just written/read to pipe.
// If you open file O_RDWR it has to start two threads!
//
// And people think I'm biased because I love developing under Unix.
//
// Cos I'm assing tired of this, it is implemented in BLOCKING mode
// as it happens localhost connects are instant anyway, But in future
// this should be change to an internal thing, using lions own
// calls to set it all up
//

//
// Perform a fork(). The child gets NULL returned to it.
// If there is a failure, a valid node is still issues, and
// the _LOST event is sent. If the child actually needs to
// get a handle as well, it can pass the "child" ptrptr.
//

//static FILE *fd = NULL;


__RCSID("$LiON: lundman/lion/pipe_win32.c,v 1.17 2006/06/30 01:22:58 lundman Exp $");





connection_t *lion_filethread(int file, int (*start_address)(void *), void *arg, void *user_data)
{
	connection_t *parent_pipe, *child_pipe;
	SECURITY_ATTRIBUTES sec;
	HANDLE shared_mutex = NULL;

	parent_pipe = connections_new();
	
	parent_pipe->type = LION_TYPE_PIPE_FILE;
	parent_pipe->user_data = user_data;
	parent_pipe->file_socket = file;

	child_pipe = connections_new();
	
	child_pipe->type = LION_TYPE_PIPE;
	child_pipe->user_data = user_data; // child should call set_userdata.
	child_pipe->file_socket = file;


#ifdef DEBUG
	printf("[thread] making pipes\n");
#endif
	// Connect the two nodes.
	if (pipe_make_pipes(parent_pipe, child_pipe))
		goto failed;

	

	// Unfortunately, we need a mutex to stop the child from running right away.

	sec.nLength = sizeof(sec);
	sec.lpSecurityDescriptor = NULL;
	sec.bInheritHandle = TRUE;

#ifdef DEBUG
	printf("[thread] creating mutex\n");
#endif
	// Create the mutex to stop the child from running.
	shared_mutex = CreateMutex( &sec, TRUE, NULL );

	if (shared_mutex == NULL)
		goto failed;
	
	child_pipe->mutex = shared_mutex;
	parent_pipe->mutex = shared_mutex;

	// We don't want to consider the child node as parent anymore
	// The child sets this to disconnect when ready.
	// This would free this node far too soon.
	//child_pipe->type = LION_TYPE_NONE;
	lion_disable_read(child_pipe);


#ifdef DEBUG
	printf("[thread] creating thread\n");
#endif

	parent_pipe->thread = CreateThread(NULL, 0, 
									   (LPTHREAD_START_ROUTINE) start_address, 
									   (void *) child_pipe,0, NULL); 


	if ( parent_pipe->thread == NULL ) {
		goto failed;
	}



	// Parent!

	// We need to release the "child_pipe" on this side as we are parent.
	// and this node no longer exists on this end.


#ifdef DEBUG
	printf("[parent] setting pending. %p\n", parent_pipe);
#endif

	// since we may or may not have data for select, force a loop to 
	// send event.
	io_force_loop = 1;
	
	parent_pipe->status = ST_PENDING;
	return parent_pipe;

	

 failed:
	// We failed. Set parent status to DISCONNECT, and release the child.
	// lnet will then issue _LOST to application
	child_pipe->status = ST_DISCONNECT;
	lion_disconnect(child_pipe);

	parent_pipe->status = ST_PENDING;
	if (parent_pipe->socket > 0)
		sockets_close(parent_pipe->socket);
	parent_pipe->socket = -1;   // This will notify it of failure, WAIT? Who closes parent pipe?

	if (shared_mutex) {
		CloseHandle(shared_mutex);
		parent_pipe->mutex = NULL;
		child_pipe->mutex = NULL;
	}

	close( file );
	// We want select to poll next iteration and enter processing
	// as we don't have a real fd to add to select, but we need to send
	// events from that part of the library
	io_force_loop = 1;

	return NULL;

}






connection_t *lion_filethread2(int file1, int file2,
							  int (*start_address1)(void *), 
							  int (*start_address2)(void *), void *user_data)
{
	connection_t *parent_pipe, *child_pipe1, *child_pipe2;
	SECURITY_ATTRIBUTES sec;
	HANDLE shared_mutex = NULL;

	parent_pipe = connections_new();
	
	parent_pipe->type = LION_TYPE_PIPE_FILE;
	parent_pipe->user_data = user_data;
	parent_pipe->file_socket = file1;

	child_pipe1 = connections_new();
	child_pipe1->type = LION_TYPE_PIPE;
	child_pipe1->user_data = user_data; // child should call set_userdata.
	child_pipe1->file_socket = file1;

	child_pipe2 = connections_new();
	child_pipe2->type = LION_TYPE_PIPE;
	child_pipe2->user_data = user_data; // child should call set_userdata.
	child_pipe2->file_socket = file2;


#ifdef DEBUG
	printf("[thread] making pipes\n");
#endif
	// Connect the two nodes.
	if (pipe_make_pipes(parent_pipe, child_pipe1))
		goto failed;

	child_pipe2->socket = child_pipe1->socket;
	

	// Unfortunately, we need a mutex to stop the child from running right away.

	sec.nLength = sizeof(sec);
	sec.lpSecurityDescriptor = NULL;
	sec.bInheritHandle = TRUE;

#ifdef DEBUG
	printf("[thread] creating mutex\n");
#endif
	// Create the mutex to stop the child from running.
	shared_mutex = CreateMutex( &sec, TRUE, NULL );

	if (shared_mutex == NULL)
		goto failed;
	
	//child_pipe1->mutex = shared_mutex;
	child_pipe2->mutex = shared_mutex;
	parent_pipe->mutex = shared_mutex;

	// We don't want to consider the child node as parent anymore
	// The child sets this to disconnect when ready.
	// This would free this node far too soon.
	//child_pipe->type = LION_TYPE_NONE;
	lion_disable_read(child_pipe1);
	lion_disable_read(child_pipe2);


#ifdef DEBUG
	printf("[thread] creating thread\n");
#endif

	parent_pipe->thread = CreateThread(NULL, 0, 
								   (LPTHREAD_START_ROUTINE) start_address1, 
									   (void *) child_pipe1,0, NULL); 


	if ( parent_pipe->thread == NULL ) {
		goto failed;
	}

	parent_pipe->thread2 = CreateThread(NULL, 0, 
								   (LPTHREAD_START_ROUTINE) start_address2, 
									   (void *) child_pipe2,0, NULL); 


	if ( parent_pipe->thread2 == NULL ) {
		goto failed;
	}


	// Parent!

	// We need to release the "child_pipe" on this side as we are parent.
	// and this node no longer exists on this end.


#ifdef DEBUG
	printf("[parent] setting pending. %p\n", parent_pipe);
#endif

	// since we may or may not have data for select, force a loop to 
	// send event.
	io_force_loop = 1;
	
	parent_pipe->status = ST_PENDING;
	return parent_pipe;

	

 failed:
	// We failed. Set parent status to DISCONNECT, and release the child.
	// lnet will then issue _LOST to application
	child_pipe1->status = ST_DISCONNECT;
	lion_disconnect(child_pipe1);
	child_pipe2->status = ST_DISCONNECT;
	lion_disconnect(child_pipe2);

	parent_pipe->status = ST_PENDING;
	if (parent_pipe->socket > 0)
		sockets_close(parent_pipe->socket);
	parent_pipe->socket = -1;   // This will notify it of failure, WAIT? Who closes parent pipe?

	if (shared_mutex) {
		CloseHandle(shared_mutex);
		parent_pipe->mutex = NULL;
		child_pipe1->mutex = NULL;
		child_pipe2->mutex = NULL;
	}

	close( file1 );
	close( file2 );
	// We want select to poll next iteration and enter processing
	// as we don't have a real fd to add to select, but we need to send
	// events from that part of the library
	io_force_loop = 1;

	return NULL;

}




int pipe_child_starter(void *child_pipe)
{
	connection_t *chpipe = (connection_t *) child_pipe;
	connection_t *newd;
	
	// Sanity checks
	if (!chpipe ||
		!chpipe->start_address ||
		!chpipe->socket)
		_exit(1);

#ifdef DEBUG
	printf("[child_starter] waiting for mutex\n");
#endif

	if (chpipe->mutex) {
		WaitForSingleObject( chpipe->mutex, INFINITE );
		CloseHandle( chpipe->mutex );
		chpipe->mutex = NULL;
	}
	// We allocate a new node here which goes into "this" thread's
	// linked-list, then we tell the other thread to release its node.

	newd = connections_new();

	connections_dupe(newd, chpipe); // to, from


	// Release the other, which is in the main threads linked list.
	chpipe->socket = -1;
	chpipe->status = ST_DISCONNECT;
	chpipe->type = LION_TYPE_NONE;  // today
	

	newd->status = ST_PENDING;  // This way the child gets a connected event.
	newd->type = LION_TYPE_PIPE;

#ifdef DEBUG
	printf("[pipe_child_starter] calling child %p %p\n", chpipe, newd);
#endif

	ExitThread(newd->start_address(newd, newd->user_data));
	return 0; // NOT REACHED, BUT MSVC++ warning avoidance.
}




connection_t *lion_fork(int (*start_address)(connection_t *, void *), 
						int flags, void *user_data, void *arg)
{
	connection_t *parent_pipe, *child_pipe;
	SECURITY_ATTRIBUTES sec;
	HANDLE shared_mutex = NULL;

	parent_pipe = connections_new();
	
	parent_pipe->type = LION_TYPE_PIPE;
	parent_pipe->user_data = user_data;

	if (flags & LION_FLAG_TRACE)
		lion_enable_trace(parent_pipe);

	if (parent_pipe->trace)
		fprintf(trace_file, "%p: lion_fork\n", parent_pipe);

	child_pipe = connections_new();
	
	child_pipe->type = LION_TYPE_PIPE;
	child_pipe->user_data = user_data; // child should call set_userdata.

#ifdef DEBUG
	printf("[thread] making pipes\n");
#endif
	// Connect the two nodes.
	if (pipe_make_pipes(parent_pipe, child_pipe))
		goto failed;
 

	// Unfortunately, we need a mutex to stop the child from running right away.

	sec.nLength = sizeof(sec);
	sec.lpSecurityDescriptor = NULL;
	sec.bInheritHandle = TRUE;

#ifdef DEBUG
	printf("[thread] creating mutex\n");
#endif
	// Create the mutex to stop the child from running.
	shared_mutex = CreateMutex( &sec, TRUE, NULL );

	if (shared_mutex == NULL)
		goto failed;
	
	child_pipe->mutex = shared_mutex;
	parent_pipe->mutex = shared_mutex;

	// We don't want to consider the child node as parent anymore
	// The child sets this to disconnect when ready.
	child_pipe->type = LION_TYPE_NONE;
	child_pipe->start_address = start_address;

#ifdef DEBUG
	printf("[thread] creating thread\n");
#endif

	parent_pipe->thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) pipe_child_starter, (void *) child_pipe,0, NULL); //&a_threadId);


	if ( parent_pipe->thread == NULL ) {
		goto failed;
	}



	// Parent!

	// We need to release the "child_pipe" on this side as we are parent.
	// and this node no longer exists on this end.


#ifdef DEBUG
	printf("[parent] setting pending. %p\n", parent_pipe);
#endif

	// since we may or may not have data for select, force a loop to 
	// send event.
	io_force_loop = 1;
	
	parent_pipe->status = ST_PENDING;

	return parent_pipe;

	

 failed:

	// We failed, so set the reason here, so we can give it to the application
	parent_pipe->error_code = errno;
	
	// We failed. Set parent status to DISCONNECT, and release the child.
	// lnet will then issue _LOST to application
	child_pipe->status = ST_DISCONNECT;
	lion_disconnect(child_pipe);

	parent_pipe->status = ST_PENDING;
	if (parent_pipe->socket > 0)
		sockets_close(parent_pipe->socket);
	parent_pipe->socket = -1;   // This will notify it of failure, WAIT? Who closes parent pipe?

	if (shared_mutex) {
		CloseHandle(shared_mutex);
		parent_pipe->mutex = NULL;
		child_pipe->mutex = NULL;
	}

	// We want select to poll next iteration and enter processing
	// as we don't have a real fd to add to select, but we need to send
	// events from that part of the library
	io_force_loop = 1;

	return NULL;

}


//
// Execute a new process. Unfortunately in windows you can not
// give a SOCKET as your STD_INPUT_HANDLE, as that takes a
// HANDLE. So here we need to create a "lion pipe" for the
// parent end. And two uni-directional pipes to be STDIN and
// STDOUT. Then start two threads:
// parent_pipe  ->  stdinWr  ->   stdin
// parent_pipe  <-  stdoutRd <-   stdout
//
connection_t *lion_execve2(char *base, char **argv, char **envp,
						  int with_stderr, int lion_flags, void *user_data)
{
	connection_t *parent_pipe, *child_pipe;
	SECURITY_ATTRIBUTES sec;
	STARTUPINFO    si;
	PROCESS_INFORMATION  pi;
    SECURITY_ATTRIBUTES sa;

	parent_pipe = connections_new();
	
	parent_pipe->type = LION_TYPE_PIPE;
	parent_pipe->user_data = user_data;

	if (lion_flags & LION_FLAG_TRACE)
		lion_enable_trace(parent_pipe);

	child_pipe = connections_new();
	
	child_pipe->type = LION_TYPE_PIPE;
	child_pipe->user_data = user_data; // child should call set_userdata.

#ifdef DEBUG
	printf("[execv] making pipes\n");
#endif

	// Connect the two nodes. This is the LION pipe.
	if (pipe_make_pipes(parent_pipe, child_pipe))
		goto failed;
 
	// Setup Security for CreatePipe.
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);     // Size of struct
    sa.lpSecurityDescriptor = NULL;               // Default 
    sa.bInheritHandle = TRUE;                    // Inheritable


	// Get ready to start the Process
	sec.nLength = sizeof(sec);
	sec.lpSecurityDescriptor = NULL;
	sec.bInheritHandle = TRUE;

	GetStartupInfo(&si);
    si.dwFlags    = STARTF_USESTDHANDLES;

	si.hStdInput  = child_pipe;
	si.hStdOutput = child_pipe;
    si.hStdError  = with_stderr ? child_pipe : GetStdHandle(STD_ERROR_HANDLE);


	// We don't want to consider the child node as parent anymore
	// The child sets this to disconnect when ready.

#ifdef DEBUG
	printf("[execv] creating process\n");
#endif

    if (!CreateProcess(
				NULL,
                argv[0],            // command line
                NULL,            // process security
                NULL,            // thread security
                TRUE,            // inherit handles-yes
#ifdef DEBUG
				0,
#else
				CREATE_NO_WINDOW,               // creation flags
#endif
				envp,            // environment block
                NULL,            // current directory
                &si,             // startup info
                &pi))            // process info (out)
		{
			goto failed;
		}

    // The Windows example closes these two HANDLES
    //CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);


	parent_pipe->thread = pi.hProcess;
	parent_pipe->pid = pi.dwProcessId;
	// Parent!
	child_pipe->socket = -1; // The new child closes it
	child_pipe->type = LION_TYPE_NONE;
	child_pipe->status = ST_DISCONNECT;

	



	// We need to release the "child_pipe" on this side as we are parent.
	// and this node no longer exists on this end.


#ifdef DEBUG
	printf("[parent] setting pending. %p\n", parent_pipe);
#endif

	// since we may or may not have data for select, force a loop to 
	// send event.
	io_force_loop = 1;
	
	parent_pipe->status = ST_PENDING;

	return parent_pipe;

	

 failed:

#ifdef DEBUG
	printf("[execv] failed %d/%d\n", errno, GetLastError());
#endif

	// We failed. Set parent status to DISCONNECT, and release the child.
	// lnet will then issue _LOST to application

	if (lion_flags & LION_FLAG_FULFILL) {
		
		//child_pipe->status = ST_DISCONNECT;
		// Close the actual socket, and set fd to -1, that way
		// we will send a FAILED event to the application.
		sockets_close(child_pipe->socket);
		child_pipe->socket = -1;


                //child_pipe->type = LION_TYPE_PIPE;
                child_pipe->error_code = ENOENT;
                child_pipe->socket = -1;
                child_pipe->status = ST_PENDING;



		// Parent side we just need to clean up quietly.
		parent_pipe->status = ST_PENDING;
		if (parent_pipe->socket > 0)
			sockets_close(parent_pipe->socket);
		parent_pipe->socket = -1;   // This will notify it of failure, WAIT? Who closes parent pipe?

		io_force_loop = 1;

		return child_pipe;
	}
	
	
	child_pipe->status = ST_DISCONNECT;
	lion_disconnect(child_pipe);

	parent_pipe->status = ST_PENDING;
	if (parent_pipe->socket > 0)
		sockets_close(parent_pipe->socket);
	parent_pipe->socket = -1;   // This will notify it of failure, WAIT? Who closes parent pipe?


	
	
	
	
	// We want select to poll next iteration and enter processing
	// as we don't have a real fd to add to select, but we need to send
	// events from that part of the library
	io_force_loop = 1;

	return NULL;

}


#define THREAD_IO_SIZE node->inbuffer_size


// Read from lion pipe (from parent) and write out to HANDLE which
// is child-process' stdin.
int lion_stdinwriter(void *arg)
{
	connection_t *node;
	int temp;
	int red = 0, wrote = 0, ret = 0;

	node = (connection_t *) arg;

	if (!node) 
			goto all_done;

	//if (fd) { fprintf(fd, " 1stdinwriter  %d <=> %d\r\n", node->file_socket, node->socket); fflush(fd);}

	// Wait for parent to release us so we don't start before they are ready.
	//WaitForSingleObject( node->mutex, INFINITE );
	//CloseHandle( node->mutex );
	//node->mutex = NULL;

	//if (fd) { fprintf(fd, " 1mutex released\r\n"); fflush(fd);}

	//printf("[lion_filewriter] Starting..\n");

	// Kick pipe into blocking IO again..
	temp = 0;
	ioctlsocket(node->socket, FIONBIO, &temp);


	while(  ( red = sockets_read(node->socket, node->buffer, THREAD_IO_SIZE) ) > 0) {

		node->buffer[red] = 0;

			//if (fd) { fprintf(fd, " 1data %d '%s'\r\n", red, node->buffer); fflush(fd);}


		//wrote = write( node->file_socket, node->buffer, red );
		//wrote = WriteFile( node->file_socket, node->buffer, red );
	    if (!WriteFile((HANDLE)node->file_socket, node->buffer, red, 
		     &wrote, NULL)) break; 

		//if (fd) { fprintf(fd, " 1sent %d\r\n", wrote); fflush(fd);}
		//printf("   io %d %d -> %d\n", red, wrote, node->socket);

		if (wrote != red) {
			break;
		}
	}



all_done:


	if (red == -1) // Error, signal failure.
		ret = WSAGetLastError();
	if (wrote == -1)
		ret = errno;
	else
		ret = GetLastError();


	// Ok, we need to cleanup. Nobody closes the file
	// so we have to here.
	close(node->file_socket);
	node->file_socket = -1;
#ifdef DEBUG
	printf("[file_win32] closed the file: status %d\n", ret);
#endif

	// The pipe will be closed, by master.
	node->status = ST_DISCONNECT;
	node->type = LION_TYPE_PIPE_FILE; // Is there a danger master see's this, before we get to exit?

	//printf("[lion_filewriter] signalling %p to be released\n", node);

	//io_force_loop = 1;
	shutdown(node->socket, 2);

	// Alas, we can not to force loop here, since that is OUR force loop, not parents.
	// So, we close the connection instead, this should wake the parent up.
	temp = node->socket;
	node->socket = -1; // Avoid the race condition.
	sockets_close(temp);

	return ret;
}



//
// This is a thread handler to read stdin IO.
//
// Once started it will wait for a mutex, which is the parent telling
// it when it is allowed to start.
// Then it will enter a loop reading from stdin, writing to the pipe.
//
int lion_stdinreader(void *arg)
{
	connection_t *node;
	int temp;
	int red = 0, wrote = 0, ret = 0;


	node = (connection_t *) arg;

	if (!node) 
			goto all_done;

	///if (fd) { fprintf(fd, " 2stdinreader  %d <=> %d\r\n", node->socket,node->file_socket ); fflush(fd);}

#ifdef DEBUG
	printf("[lion_stdinreader] waiting permission to start!\n");
#endif

	WaitForSingleObject( node->mutex, INFINITE );
	CloseHandle( node->mutex );
	node->mutex = NULL;

	//if (fd) { fprintf(fd, " 2mutex released\r\n"); fflush(fd);}


#ifdef DEBUG
	printf("[lion_stdinreader] Starting..\n");
#endif
	// Kick pipe into blocking IO again..
	temp = 0;
	ioctlsocket(node->socket, FIONBIO, &temp);



	while(  ReadFile((HANDLE)node->file_socket, node->buffer, THREAD_IO_SIZE, &red, NULL) ) {


		if (red <=0) break;

		node->buffer[red] = 0;

		wrote = sockets_write( node->socket, node->buffer, red );

		// Why does this thread call lion?
		lion_poll(0, 0);  // the read is blocking, so just go as fast as possible!

		if (wrote != red) {
			break;
		}
	}



all_done:

	if (!ret) 
		ret = GetLastError();
	else
		ret = errno;

	if (wrote == -1)
		ret = WSAGetLastError();

	// Ok, we need to cleanup. Nobody closes the file
	// so we have to here.
	CloseHandle((HANDLE)node->file_socket);
	node->file_socket = -1;
#ifdef DEBUG
	printf("[file_win32] closed the file: status %d\n", ret);
#endif

	// The pipe will be closed, by master.
	node->status = ST_DISCONNECT;
	node->type = LION_TYPE_PIPE_FILE; // Is there a danger master see's this, before we get to exit?

	//printf("[lion_filereader] signalling %p to be released\n", node);

	//io_force_loop = 1;
	// Alas, we can not to force loop here, since that is OUR force loop, not parents.
	// So, we close the connection instead, this should wake the parent up.

	// We need to wake up the write thread too.. So close
	// socket HARD.
	shutdown(node->socket, 2);

	temp = node->socket;
	node->socket = -1; // Avoid the race condition.
	sockets_close(temp);

	return ret;
}



connection_t *lion_execve(char *base, char **argv, char **envp,
						  int with_stderr, int lion_flags, void *user_data)
{
	connection_t *parent_pipe = NULL, *child_pipe = NULL, *newd = NULL;
	SECURITY_ATTRIBUTES sec;
	STARTUPINFO    si;
	PROCESS_INFORMATION  pi;
    SECURITY_ATTRIBUTES sa;
    HANDLE hChildStdinRd = 0, hChildStdinWr = 0,  
           hChildStdoutRd = 0, hChildStdoutWr = 0;
	int i, len = 0;
	char *cmd = NULL;

	//fd = fopen("c:/tmp/yes.txt", "a");
	//if (fd) { fprintf(fd, "execve started '%s' '%s'\r\n", argv[0], argv[1]); fflush(fd);}

	// Windows CreateProcess does not take argv vector, so we need to make
	// it into a string again, so we are POSIX compliant.
	for (i = 0; argv[i]; i++) 
		len += strlen(argv[i]) + 1;  // +space OR +null

	if (len) {
		cmd = (char *) malloc(len);
		if (!cmd) goto failed;
		*cmd = 0;
		for (i = 0; argv[i]; i++) {
			strcat(cmd, argv[i]);
			if (argv[i+1])
				strcat(cmd, " ");
		}
	} else
		cmd = strdup(base);
	if (!cmd) goto failed;

	//if (fd) { fprintf(fd, "base '%s'  cmd '%s'\r\n", base, cmd); fflush(fd);}
	


	newd = connections_new();
	
	newd->type = LION_TYPE_PIPE;
	newd->user_data = user_data;

	if (lion_flags & LION_FLAG_TRACE)
		lion_enable_trace(newd);

	if (newd->trace)
		fprintf(trace_file, "%p: lion_execv(%s)\n", base);


	// Setup Security for CreatePipe.
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);     // Size of struct
    sa.lpSecurityDescriptor = NULL;               // Default 
    sa.bInheritHandle = TRUE;                    // Inheritable

	// Create STDIN and STDOUT
    if (! CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &sa, 0)) 
        goto failed;

	if (! CreatePipe(&hChildStdinRd, &hChildStdinWr, &sa, 0)) 
        goto failed;

    //si.hStdError  = with_stderr ? hChildStdoutWr : GetStdHandle(STD_ERROR_HANDLE);


	// We don't want to consider the child node as parent anymore
	SetHandleInformation( hChildStdoutWr, HANDLE_FLAG_INHERIT, 0);
	SetHandleInformation( hChildStdinRd, HANDLE_FLAG_INHERIT, 0);

	SetHandleInformation( hChildStdinWr, HANDLE_FLAG_INHERIT, 1);
	SetHandleInformation( hChildStdoutRd, HANDLE_FLAG_INHERIT, 1);
	parent_pipe = lion_filethread2((int)hChildStdinWr, 
								   (int)hChildStdoutRd,
								   lion_stdinwriter, 
								   lion_stdinreader, 
								   user_data);
	if (!parent_pipe) goto failed;

	//CloseHandle(hChildStdinWr);
	//CloseHandle(hChildStdoutRd);

#ifdef DEBUG
	printf("[execv] creating process\n");
#endif
	//if (fd) { fprintf(fd, "launching process\r\n"); fflush(fd);}
	SetHandleInformation( hChildStdoutWr, HANDLE_FLAG_INHERIT, 1);
	SetHandleInformation( hChildStdinRd, HANDLE_FLAG_INHERIT, 1);
	SetHandleInformation( hChildStdoutRd, HANDLE_FLAG_INHERIT, 0);
	SetHandleInformation( hChildStdinWr, HANDLE_FLAG_INHERIT, 0);

	// Get ready to start the Process
	sec.nLength = sizeof(sec);
	sec.lpSecurityDescriptor = NULL;
	sec.bInheritHandle = TRUE;

	GetStartupInfo(&si);
    si.dwFlags    = STARTF_USESTDHANDLES;

	si.hStdInput  = hChildStdinRd;
	si.hStdOutput = hChildStdoutWr;

    if (!CreateProcess(
				NULL,
                cmd,             // command line
                NULL,            // process security
                NULL,            // thread security
                TRUE,            // inherit handles-yes
#ifdef DEBUG
				0,
#else
				CREATE_NO_WINDOW,               // creation flags
#endif
                envp,            // environment block
                NULL,            // current directory
                &si,             // startup info
                &pi))            // process info (out)
		{
			goto failed;
		}


	SAFE_FREE(cmd);
    // The Windows example closes these two HANDLES
    //CloseHandle(pi.hProcess);
	//WaitForInputIdle(pi.hProcess, 1000);

    CloseHandle(pi.hThread);
	CloseHandle(hChildStdinRd);
	CloseHandle(hChildStdoutWr);

	// Close child's end of pipes (since we are parent).

	//if (fd) { fprintf(fd, "calling thread2\r\n"); fflush(fd);}
	// Start a new thread to handle stdin. 
	// reads from newd->socket -> newd->file_socket.

	//if (fd) { fprintf(fd, "cleaning up\r\n"); fflush(fd);}

	parent_pipe->type = LION_TYPE_PIPE;

	// Assign pid.
	parent_pipe->thread = pi.hProcess;
	parent_pipe->pid = pi.dwProcessId;
	parent_pipe->file_socket = hChildStdinWr;
	parent_pipe->binary = newd->binary;
	parent_pipe->user_data = user_data;

	// We only want to use one bi-directional pipe in lion
	// so close the 2nd, and assign over the first. This is
	// ok because both threads are waiting on mutex. Once they
	// start they will both use ->socket, which is the same value.

	newd->socket = -1;
	newd->status = ST_DISCONNECT;  // will be released eventually.

#ifdef DEBUG
	printf("[parent] setting pending. %p\n", parent_pipe);
#endif


	// since we may or may not have data for select, force a loop to 
	// send event.
	io_force_loop = 1;
	
	parent_pipe->status = ST_PENDING;


	return parent_pipe;







 failed:

	//if (fd) { fprintf(fd, "failed!\r\n"); fflush(fd);}

#ifdef DEBUG
	printf("[execv] failed %d/%d\n", errno, GetLastError());
#endif
	SAFE_FREE(cmd);
	// We failed. Set parent status to DISCONNECT, and release the child.
	// lnet will then issue _LOST to application


	
		//if (fd) { fprintf(fd, "freeing\r\n"); fflush(fd);}
#if 1
		if (child_pipe) {
			child_pipe->status = ST_DISCONNECT;
			lion_disconnect(child_pipe);
		}
		if (parent_pipe) {
			parent_pipe->status = ST_PENDING;
			if (parent_pipe->socket > 0)
				sockets_close(parent_pipe->socket);
			parent_pipe->socket = -1;   // This will notify it of failure, WAIT? Who closes parent pipe?
		}
#endif

	if (hChildStdinRd) {
		CloseHandle(hChildStdinRd);
		hChildStdinRd = NULL;
	}
	if (hChildStdinWr) {
		CloseHandle(hChildStdinWr);
		hChildStdinWr = NULL;
	}
	if (hChildStdoutRd) {
		CloseHandle(hChildStdoutRd);
		hChildStdoutRd = NULL;
	}
	if (hChildStdoutWr) {
		CloseHandle(hChildStdoutWr);
		hChildStdoutWr = NULL;
	}


	newd->socket = -1;
	newd->status = ST_DISCONNECT;  // will be released eventually.
	
	// We want select to poll next iteration and enter processing
	// as we don't have a real fd to add to select, but we need to send
	// events from that part of the library
	io_force_loop = 1;
	//if (fd) { fprintf(fd, "returnring\r\n"); fflush(fd);}
	
	if (lion_flags & LION_FLAG_FULFILL) {
		newd->status = ST_PENDING;  // will be released eventually.
		newd->error_code = ENOENT;
		return newd;
	}

	return NULL;

}








//
// Connect two handles with pipes
//
int      pipe_make_pipes(connection_t *node1, connection_t *node2)
{
	int fd[2];
	// This is assuming you have socketpair function, and it has the UNIX
	// domain!

	// In future, we could also implement this by using our own functions
	// to open a listen on localhost, and connect to it, and tie the
	// two connections together.
	if( sockets_socketpair(fd) < 0 ) return -1;

	node1->socket = fd[0];
	node2->socket = fd[1];
	return 0; // Failure

}


int pipe_get_status( connection_t *node)
{
	int r;

#ifdef DEBUG
	printf("win32 get status: %p\n", node->thread);
#endif

	// Ok we don't know it yet.
	if (node->thread) {
		GetExitCodeThread( node->thread, &r );
		//printf("exit code %d\n", r);
		return r;
	}

	return -1;
}



void lion_want_returncode( connection_t *node)
{

	node->status = ST_WANTRETURN;

}



static char *argv[MAX_ARG];

char **parser(char *s)
{
        char *ar, *arg;
        int argc;
        static char name[1024];

        if (!s) return NULL;

        strncpy(name, s, sizeof(name));
        //name = strdup(s);
        ar   = name;


        // For each argument, build a list.
        for( argc = 0; 
                 (arg = misc_digtoken(&ar, " \t\r\n")) && (argc < MAX_ARG-1);
                 argc++ ) {

                printf("  '%s'\n", arg);

                argv[ argc ] = arg;

        }

        // null-terminate.
        argv[ argc ] = NULL;

#ifdef DEBUG
        printf("[parse] %d arguments\n", argc);
#endif

        return argv;

}



connection_t *lion_system(char *name_orig, int with_stderr,
                          int lion_flags, void *user_data)
{
	   connection_t *ret;
		char **argv;
        
        // Since the user might have passed us a static-string, which we can't
        // modify, we need to make a copy of it, and release.
        if (!name_orig) return NULL;


        argv = parser( name_orig );

		if (lion_flags & LION_FLAG_STDIO_AS_LION)
			ret = lion_execve2(argv[0], argv, NULL,
							with_stderr, lion_flags, user_data);
		else
			ret = lion_execve(argv[0], argv, NULL,
							with_stderr, lion_flags, user_data);
        return ret;

}




//
// Start a new process, leaving a pipe to the process.
// We fork, then exec a new process.
//
#if 0
connection_t *lion_execv(char *base, char **argv, int with_stderr,
						void *user_data)
{
	connection_t *parent = NULL; // the node that we give to parent!
	connection_t *child  = NULL;
	int i, err;

	// NULL means child.

//	if ((parent = lion_fork( &child, user_data ))) {
	{
		// Parent code!
		
		// Lets have a sneak peak, did the fork work at all?
		if (parent->socket == -1) {
			// fork failed - so we just exit, the event gets posted
			// later.

			return parent;

		}

		// fork worked (but the execv might still fail before parents
		// gets here! )
		
		return parent;

	}




	/* child code! */

	// Set stdin|out and possibly err to the pipe.

    dup2(child->socket, 0);
    dup2(child->socket, 1);

	if (with_stderr) {
		dup2(child->socket, 2);
	} else {
		i = open("/dev/null", O_WRONLY);
		dup2(i, 2);
		close(i);
	}
	close(child->socket);
	child->socket = 0;


	// Now attempt to start the new process
    
    err = execv(argv[0], argv);

	// If we reach this place, execv failed.
	
	child->status = ST_DISCONNECT; // dont post events
	lion_disconnect( child );       // release the only node left.

    _exit(err);

}



//#if 0



connection_t *lion_system(char *name_orig, int with_stderr,
						 void *user_data)
{
	int argc;
	char *cp, *argv[100]; 
	char *name;
	connection_t *ret;
	
	// Since the user might have passed us a static-string, which we can't
	// modify, we need to make a copy of it, and release.
	if (!name_orig) return NULL;

	name = strdup(name_orig);
	
	for (argc = 0, cp = name;; cp = NULL) {
		if (!(argv[argc++] = strtok(cp, " \t\n")))
			break;
		
		
		/*
		 * This could win the World Prize for aweful code! :)
		 */
		
		if (argv[argc-1][ strlen(argv[argc-1])+1 ] == '"') { 
			/* Ah next token has " to start with. */
			char *r;
			
			
			r = &(argv[argc-1][ strlen(argv[argc-1]) + 2 ]);
			
			cp = r;
			
			if (!(argv[argc++] = strtok(cp, "\"\t\n")))
				break;
			
			cp = NULL;
			
		}
	}
	
	if (!argv[argc - 1]) argc--;


	ret = lion_execv(argv[0], argv, with_stderr, user_data);

	free(name);

	return ret;

}









#endif



#endif

